import os
from datetime import datetime, timedelta
from pathlib import Path
from feathr import (BOOLEAN, FLOAT, INPUT_CONTEXT, INT32, STRING,
                    DerivedFeature, Feature, FeatureAnchor, HdfsSource,
                    TypedKey, ValueType, WindowAggTransformation)
from feathr import FeathrClient
from feathr.definition.sink import CosmosDbSink, ElasticSearchSink
from feathr.definition.source import HdfsSource

import pytest
from click.testing import CliRunner

from feathr import (BackfillTime, MaterializationSettings)
from feathr import FeathrClient
from feathr import FeatureQuery
from feathr import ObservationSettings
from feathr import RedisSink, HdfsSink, JdbcSink,AerospikeSink
from feathr import TypedKey
from feathr import ValueType
from feathr.utils.job_utils import get_result_df
from feathrcli.cli import init
from test_fixture import (basic_test_setup, get_online_test_table_name, composite_keys_test_setup)
from test_utils.constants import Constants

# make sure you have run the upload feature script before running these tests
# the feature configs are from feathr_project/data/feathr_user_workspace
def test_feathr_materialize_to_offline():
    """
    Test FeathrClient() HdfsSink.
    """
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    # os.chdir(test_workspace_dir)

    client: FeathrClient = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    backfill_time = BackfillTime(start=datetime(
        2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))

    now = datetime.now()
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob_materialize_offline_','_', str(now.minute), '_', str(now.second), ""])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/feathrazure_cijob_materialize_offline_','_', str(now.minute), '_', str(now.second), ""])
    offline_sink = HdfsSink(output_path=output_path)
    settings = MaterializationSettings("nycTaxiTable",
                                       sinks=[offline_sink],
                                       feature_names=[
                                           "f_location_avg_fare", "f_location_max_fare"],
                                       backfill_time=backfill_time)
    client.materialize_features(settings)
    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    # download result and just assert the returned result is not empty
    # by default, it will write to a folder appended with date
    res_df = get_result_df(client, data_format="avro", res_url=output_path + "/df0/daily/2020/05/20")
    assert res_df.shape[0] > 0

def test_feathr_online_store_agg_features():
    """
    Test FeathrClient() get_online_features and batch_get can get data correctly.
    """

    online_test_table = get_online_test_table_name("nycTaxiCITableSparkE2E")
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    # os.chdir(test_workspace_dir)

    client: FeathrClient = composite_keys_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    backfill_time = BackfillTime(start=datetime(
        2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))
    redisSink = RedisSink(table_name=online_test_table)
    settings = MaterializationSettings("nycTaxiTable",
                                       sinks=[redisSink],
                                       feature_names=[
                                           "f_location_avg_fare", "f_location_max_fare"],
                                       backfill_time=backfill_time)
    client.materialize_features(settings)
    # just assume the job is successful without validating the actual result in Redis. Might need to consolidate
    # this part with the test_feathr_online_store test case
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    res = client.get_online_features(online_test_table, ["81", "254"], [
                                     'f_location_avg_fare', 'f_location_max_fare'])
    # just assume there are values. We don't hard code the values for now for testing
    # the correctness of the feature generation should be guaranteed by feathr runtime.
    # ID 239 and 265 are available in the `DOLocationID` column in this file:
    # https://s3.amazonaws.com/nyc-tlc/trip+data/green_tripdata_2020-04.csv
    # View more details on this dataset: https://www1.nyc.gov/site/tlc/about/tlc-trip-record-data.page
    assert res != None
    res = client.multi_get_online_features(online_test_table,
                                           [["81","254"], ["25","42"]],
                                           ['f_location_avg_fare', 'f_location_max_fare'])
    assert res['81#254'] != None
    assert res['25#42'] != None
    
    client._clean_test_data(online_test_table)

@pytest.mark.skip(reason="Add back when complex types are supported in python API")
def test_feathr_online_store_non_agg_features():
    """
    Test FeathrClient() online_get_features and batch_get can get data correctly.
    """
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    client = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    online_test_table = get_online_test_table_name('nycTaxiCITableNonAggFeature')
    backfill_time = BackfillTime(start=datetime(
        2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))
    redisSink = RedisSink(table_name=online_test_table)
    settings = MaterializationSettings("nycTaxiTable",
                                       sinks=[redisSink],
                                       feature_names=["f_gen_trip_distance", "f_gen_is_long_trip_distance", "f1", "f2", "f3", "f4", "f5", "f6"],
                                       backfill_time=backfill_time)

    client.materialize_features(settings, allow_materialize_non_agg_feature=True)
    # just assume the job is successful without validating the actual result in Redis. Might need to consolidate
    # this part with the test_feathr_online_store test case
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    res = client.get_online_features(online_test_table, '111', ['f_gen_trip_distance', 'f_gen_is_long_trip_distance',
                                                                   'f1', 'f2', 'f3', 'f4', 'f5', 'f6'])
    # just assume there are values. We don't hard code the values for now for testing
    # the correctness of the feature generation should be guaranteed by feathr runtime.
    # ID 239 and 265 are available in the `DOLocationID` column in this file:
    # https://s3.amazonaws.com/nyc-tlc/trip+data/green_tripdata_2020-04.csv
    # View more detail on this dataset: https://www1.nyc.gov/site/tlc/about/tlc-trip-record-data.page

    assert len(res) == 8
    assert res[0] != None
    assert res[1] != None
    # assert constant features
    _validate_constant_feature(res)
    res = client.multi_get_online_features(online_test_table,
                                           ['239', '265'],
                                           ['f_gen_trip_distance', 'f_gen_is_long_trip_distance', 'f1', 'f2', 'f3', 'f4', 'f5', 'f6'])
    _validate_constant_feature(res['239'])
    assert res['239'][0] != None
    assert res['239'][1] != None
    _validate_constant_feature(res['265'])
    assert res['265'][0] != None
    assert res['265'][1] != None
    
    client._clean_test_data(online_test_table)


def _validate_constant_feature(feature):
    assert feature[2] == [10.0, 20.0, 30.0]
    assert feature[3] == ['a', 'b', 'c']
    assert feature[4] == ([1, 2, 3], ['10', '20', '30'])
    assert feature[5] == ([1, 2, 3], [True, False, True])
    assert feature[6] == ([1, 2, 3], [1.0, 2.0, 3.0])
    assert feature[7] == ([1, 2, 3], [1, 2, 3])


def test_dbfs_path():
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    client = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    if client.spark_runtime.casefold() == "databricks":
        # expect this raise an error since the result path is not in dbfs: format
        with pytest.raises(RuntimeError):
            client.feathr_spark_launcher.download_result(result_path="wasb://res_url", local_folder="/tmp")


def test_feathr_get_offline_features():
    """
    Test get_offline_features() can get data correctly.
    """
    runner = CliRunner()
    with runner.isolated_filesystem():
        runner.invoke(init, [])
        client = basic_test_setup(
            "./feathr_user_workspace/feathr_config.yaml")

        location_id = TypedKey(key_column="DOLocationID",
                               key_column_type=ValueType.INT32,
                               description="location id in NYC",
                               full_name="nyc_taxi.location_id")

        feature_query = FeatureQuery(
            feature_list=["f_location_avg_fare", "f_trip_time_rounded"], key=location_id)
        settings = ObservationSettings(
            observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
            event_timestamp_column="lpep_dropoff_datetime",
            timestamp_format="yyyy-MM-dd HH:mm:ss")

        now = datetime.now()
        # set output folder based on different runtime
        if client.spark_runtime == 'databricks':
            output_path = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), ".avro"])
        else:
            output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/output','_', str(now.minute), '_', str(now.second), ".avro"])


        client.get_offline_features(observation_settings=settings,
                                    feature_query=feature_query,
                                    output_path=output_path)

        # assuming the job can successfully run; otherwise it will throw exception
        client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

        # download result and just assert the returned result is not empty
        res_df = get_result_df(client)
        assert res_df.shape[0] > 0

def test_feathr_get_offline_features_to_sql():
    """
    Test get_offline_features() can save data to SQL.
    """
    # runner.invoke(init, [])
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    client: FeathrClient = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    location_id = TypedKey(key_column="DOLocationID",
                            key_column_type=ValueType.INT32,
                            description="location id in NYC",
                            full_name="nyc_taxi.location_id")

    feature_query = FeatureQuery(
        feature_list=["f_location_avg_fare"], key=location_id)
    settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")

    now = datetime.now()
    # Set DB user and password before submitting job
    # os.environ[f"sql1_USER"] = "some_user@feathrtestsql4"
    # os.environ[f"sql1_PASSWORD"] = "some_password"
    output_path = JdbcSink(name="sql1",
                            url="jdbc:sqlserver://feathrazureci.database.windows.net:1433;database=feathrci;encrypt=true;",
                            dbtable=f'feathr_ci_materialization_{str(now)[:19].replace(" ", "_").replace(":", "_").replace("-", "_")}',
                            auth="USERPASS")

    client.get_offline_features(observation_settings=settings,
                                feature_query=feature_query,
                                output_path=output_path)

    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

@pytest.mark.skip(reason="Marked as skipped as we need to setup token and enable SQL AAD login for this test")
def test_feathr_get_offline_features_to_sql_with_token():
    """
    Test get_offline_features() can save data to SQL.
    """
    # runner.invoke(init, [])
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    client: FeathrClient = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    location_id = TypedKey(key_column="DOLocationID",
                            key_column_type=ValueType.INT32,
                            description="location id in NYC",
                            full_name="nyc_taxi.location_id")

    feature_query = FeatureQuery(
        feature_list=["f_location_avg_fare"], key=location_id)
    settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")

    now = datetime.now()

    # Set DB token before submitting job
    # os.environ[f"SQL1_TOKEN"] = "some_token"
    os.environ["SQL1_TOKEN"] = client.credential.get_token("https://management.azure.com/.default").token
    output_path = JdbcSink(name="sql1",
                            url="jdbc:sqlserver://feathrazureci.database.windows.net:1433;database=feathrci;encrypt=true;",
                            dbtable=f'feathr_ci_sql_token_{str(now)[:19].replace(" ", "_").replace(":", "_").replace("-", "_")}',
                            auth="TOKEN")

    client.get_offline_features(observation_settings=settings,
                                feature_query=feature_query,
                                output_path=output_path)

    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

@pytest.mark.skipif(os.environ.get('SPARK_CONFIG__SPARK_CLUSTER') == "databricks",
                    reason="Due to package conflicts, the CosmosDB test doesn't work on databricks clusters, refer to https://github.com/feathr-ai/feathr/blob/main/docs/how-to-guides/jdbc-cosmos-notes.md#using-cosmosdb-as-the-online-store for more details")
def test_feathr_materialize_to_cosmosdb():
    """
    Test FeathrClient() CosmosDbSink.
    """
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    # os.chdir(test_workspace_dir)

    client: FeathrClient = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    backfill_time = BackfillTime(start=datetime(
        2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))

    now = datetime.now()
    container = ''.join(['feathrazure_cijob_materialize_','_', str(now.minute), '_', str(now.second), ""])
    sink = CosmosDbSink(name='cosmos1', endpoint='https://feathrazuretest3-cosmosdb.documents.azure.com:443/', database='feathr', container=container)
    settings = MaterializationSettings("nycTaxiTable",
                                       sinks=[sink],
                                       feature_names=[
                                           "f_location_avg_fare", "f_location_max_fare"],
                                       backfill_time=backfill_time)
    client.materialize_features(settings)
    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

@pytest.mark.skip(reason="Marked as skipped as we need to setup resources for this test")
def test_feathr_materialize_to_es():
    """
    Test FeathrClient() CosmosDbSink.
    """
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    # os.chdir(test_workspace_dir)

    client: FeathrClient = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    backfill_time = BackfillTime(start=datetime(
        2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))

    now = datetime.now()
    index = ''.join(['feathrazure_cijob_materialize_','_', str(now.minute), '_', str(now.second), ""])
    sink = ElasticSearchSink(name='es1', host='somenode:9200', index=index, ssl=True, auth=True)
    settings = MaterializationSettings("nycTaxiTable",
                                       sinks=[sink],
                                       feature_names=[
                                           "f_location_avg_fare", "f_location_max_fare"],
                                       backfill_time=backfill_time)
    client.materialize_features(settings)
    # Set user and password before submitting job
    # os.environ[f"es1_USER"] = "some_user"
    # os.environ[f"es1_PASSWORD"] = "some_password"
    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

@pytest.mark.skip(reason="Marked as skipped as we need to setup resources for this test")
def test_feathr_materialize_to_aerospike():
    """
    Test FeathrClient() CosmosDbSink.
    """
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    # os.chdir(test_workspace_dir)
    now = datetime.now()
    # set workspace folder by time; make sure we don't have write conflict if there are many CI tests running
    os.environ['SPARK_CONFIG__DATABRICKS__WORK_DIR'] = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), '_', str(now.microsecond)])
    os.environ['SPARK_CONFIG__AZURE_SYNAPSE__WORKSPACE_DIR'] = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/feathr_github_ci','_', str(now.minute), '_', str(now.second) ,'_', str(now.microsecond)])

    client = FeathrClient(config_path="feathr_config.yaml")
    batch_source = HdfsSource(name="nycTaxiBatchSource",
                              path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    f_trip_distance = Feature(name="f_trip_distance",
                              feature_type=FLOAT, transform="trip_distance")
    f_trip_time_duration = Feature(name="f_trip_time_duration",
                                   feature_type=INT32,
                                   transform="(to_unix_timestamp(lpep_dropoff_datetime) - to_unix_timestamp(lpep_pickup_datetime))/60")

    features = [
        f_trip_distance,
        f_trip_time_duration,
        Feature(name="f_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>30"),
        Feature(name="f_day_of_week",
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]


    request_anchor = FeatureAnchor(name="request_features",
                                   source=INPUT_CONTEXT,
                                   features=features)

    f_trip_time_distance = DerivedFeature(name="f_trip_time_distance",
                                          feature_type=FLOAT,
                                          input_features=[
                                              f_trip_distance, f_trip_time_duration],
                                          transform="f_trip_distance * f_trip_time_duration")

    f_trip_time_rounded = DerivedFeature(name="f_trip_time_rounded",
                                         feature_type=INT32,
                                         input_features=[f_trip_time_duration],
                                         transform="f_trip_time_duration % 10")

    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")
    agg_features = [Feature(name="avgfare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="AVG",
                                                              window="90d",
                                                              )),
                    Feature(name="maxfare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="MAX",
                                                              window="90d"))
                    ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=batch_source,
                               features=agg_features)

    client.build_features(anchor_list=[agg_anchor, request_anchor], derived_feature_list=[
        f_trip_time_distance, f_trip_time_rounded])


    backfill_time = BackfillTime(start=datetime(
        2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))

    now = datetime.now()
    os.environ[f"aerospike_USER"] = "feathruser"
    os.environ[f"aerospike_PASSWORD"] = "feathr"
    as_sink = AerospikeSink(name="aerospike",seedhost="20.57.186.153", port=3000, namespace="test", setname="test")
    settings = MaterializationSettings("nycTaxiTable",
                                       sinks=[as_sink],
                                       feature_names=[
                                           "avgfare", "maxfare"],
                                       backfill_time=backfill_time)
    client.materialize_features(settings)
    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

if __name__ == "__main__":
    test_feathr_materialize_to_aerospike()
    test_feathr_get_offline_features_to_sql()
    test_feathr_materialize_to_cosmosdb()




import glob
import os
import time
import unittest
import pytest
from datetime import datetime
from pathlib import Path

from click.testing import CliRunner

from feathr import (FeatureQuery, ObservationSettings, TypedKey,
                    ValueType)
from feathr.client import FeathrClient
from feathr.registry._feathr_registry_client import _FeatureRegistry
from feathrcli.cli import init
from test_fixture import registry_test_setup
from test_fixture import registry_test_setup_append, registry_test_setup_partially, registry_test_setup_for_409
from test_utils.constants import Constants

class FeatureRegistryTests(unittest.TestCase):
    def test_feathr_register_features_e2e(self):
        """
        This test will register features, get all the registered features, then query a set of already registered features.
        """

        config_paths = [
            "feathr_config_registry_purview.yaml",
            "feathr_config_registry_purview_rbac.yaml",
            "feathr_config_registry_sql.yaml",
            "feathr_config_registry_sql_rbac.yaml",
        ]

        for config_path in config_paths:
            with self.subTest(config_path=config_path):
                test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"
                client: FeathrClient = registry_test_setup(os.path.join(test_workspace_dir, config_path))

                # set output folder based on different runtime
                now = datetime.now()
                if client.spark_runtime == 'databricks':
                    output_path = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), ".parquet"])
                else:
                    output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/output','_', str(now.minute), '_', str(now.second), ".parquet"])


                client.register_features()
                # Allow purview to process a bit
                time.sleep(5)
                # in CI test, the project name is set by the CI pipeline so we read it here
                all_features = client.list_registered_features(project_name=client.project_name)
                all_feature_names = [x['name'] for x in all_features]

                assert 'f_is_long_trip_distance' in all_feature_names # test regular ones
                assert 'f_trip_time_rounded' in all_feature_names # make sure derived features are there
                assert 'f_location_avg_fare' in all_feature_names # make sure aggregated features are there
                assert 'f_trip_time_rounded_plus' in all_feature_names # make sure derived features are there
                assert 'f_trip_time_distance' in all_feature_names # make sure derived features are there

                # Sync workspace from registry, will get all conf files back
                client.get_features_from_registry(client.project_name)
                
                # Register the same feature with different definition and expect an error.
                client: FeathrClient = registry_test_setup_for_409(os.path.join(test_workspace_dir, config_path), client.project_name)

                with pytest.raises(RuntimeError) as exc_info:
                    client.register_features()
                
                # <ExceptionInfo RuntimeError('Failed to call registry API, status is 409, error is {"message":"Entity feathr_ci_registry_53_3_476999__request_features__f_is_long_trip_distance already exists"}')
            assert "status is 409" in str(exc_info.value)

    def test_feathr_register_features_partially(self):
        """
        This test will register full set of features into one project, then register another project in two partial registrations.
        The length of the return value of get_features_from_registry should be identical.
        """
        test_workspace_dir = Path(
            __file__).parent.resolve() / "test_user_workspace"
        client: FeathrClient = registry_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))
        client.register_features()
        time.sleep(30)
        full_registration, keys = client.get_features_from_registry(client.project_name, return_keys = True, verbose = True)
        assert len(keys['f_location_avg_fare']) == 2
        now = datetime.now()
        os.environ["project_config__project_name"] =  ''.join(['feathr_ci_registry','_', str(now.minute), '_', str(now.second), '_', str(now.microsecond)])

        client: FeathrClient = registry_test_setup_partially(os.path.join(test_workspace_dir, "feathr_config.yaml"))
        new_project_name = client.project_name
        client.register_features()
        time.sleep(30)


        client: FeathrClient = registry_test_setup_append(os.path.join(test_workspace_dir, "feathr_config.yaml"))
        client.project_name = new_project_name
        client.register_features()
        time.sleep(30)

        appended_registration = client.get_features_from_registry(client.project_name)

        # after a full registration, another registration should not affect the registered anchor features.
        assert len(full_registration.items())==len(appended_registration.items())

    @pytest.mark.skip(reason="Underlying implementation changed, not applicable")
    def test_get_feature_from_registry(self):
        registry = _FeatureRegistry("mock_project","mock_purview","mock_delimeter")
        derived_feature_with_multiple_inputs = {
                "guid": "derived_feature_with_multiple_input_anchors",
                "typeName": "feathr_derived_feature_v1",
                "attributes": {
                    "input_derived_features": [],
                    "input_anchor_features": [
                        {
                            "guid": "input_anchorA",
                            "typeName": "feathr_anchor_feature_v1",
                        },
                        {
                            "guid": "input_anchorB",
                            "typeName": "feathr_anchor_feature_v1",
                        }
                    ]
                },
            }
        hierarchical_derived_feature = {
                "guid": "hierarchical_derived_feature",
                "typeName": "feathr_derived_feature_v1",
                "attributes": {
                    "input_derived_features": [
                        {
                            "guid": "derived_feature_with_multiple_input_anchors",
                            "typeName": "feathr_derived_feature_v1",
                        }
                    ],
                    "input_anchor_features": [
                        {
                            "guid": "input_anchorC",
                            "typeName": "feathr_anchor_feature_v1",
                        }
                    ],
                }
            }
        anchors = [
            {
                "guid": "input_anchorA",
                "typeName": "feathr_anchor_feature_v1",
            },
            {
                "guid": "input_anchorC",
                "typeName": "feathr_anchor_feature_v1",
            },
            {
                "guid": "input_anchorB",
                "typeName": "feathr_anchor_feature_v1",
            }]

        def entity_array_to_dict(arr):
            return {x['guid']:x for x in arr}

        inputs = registry.search_input_anchor_features(['derived_feature_with_multiple_input_anchors'],entity_array_to_dict(anchors+[derived_feature_with_multiple_inputs]))
        assert len(inputs)==2
        assert "input_anchorA" in inputs and "input_anchorB" in inputs

        inputs = registry.search_input_anchor_features(['hierarchical_derived_feature'],entity_array_to_dict(anchors+[derived_feature_with_multiple_inputs,hierarchical_derived_feature]))
        assert len(inputs)==3
        assert "input_anchorA" in inputs and "input_anchorB" in inputs and "input_anchorC" in inputs

    @pytest.mark.skip(reason="Add back get_features is not supported in feature registry for now and needs further discussion")
    def test_feathr_get_features_from_registry(self):
        """
        Test FeathrClient() sync features and get all the conf files from registry
        """
        runner = CliRunner()
        with runner.isolated_filesystem():
            result = runner.invoke(init, [])

            assert result.exit_code == 0
            assert os.path.isdir("./feathr_user_workspace")
            os.chdir('feathr_user_workspace')

            # Look for conf files, we shouldn't have any
            total_conf_files = glob.glob('*/*.conf', recursive=True)
            assert len(total_conf_files) == 0

            client = FeathrClient()
            # Sync workspace from registry, will get all conf files back
            client.get_features_from_registry("frame_getting_started")

            total_conf_files = glob.glob('*/*.conf', recursive=True)
            # we should have at least 3 conf files
            assert len(total_conf_files) == 3





from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Union, Set

from loguru import logger
import pandas as pd
import re
from pyspark.sql import DataFrame, SparkSession

from feathr.client import FeathrClient
from feathr.constants import OUTPUT_FORMAT
from feathr.utils.platform import is_databricks
from feathr.spark_provider._synapse_submission import _DataLakeFiler

def get_result_pandas_df(
    client: FeathrClient,
    data_format: str = None,
    res_url: str = None,
    local_cache_path: str = None,
) -> pd.DataFrame:
    """Download the job result dataset from cloud as a Pandas DataFrame.

    Args:
        client: Feathr client
        data_format: Format to read the downloaded files. Currently support `parquet`, `delta`, `avro`, and `csv`.
            Default to use client's job tags if exists.
        res_url: Result URL to download files from. Note that this will not block the job so you need to make sure
            the job is finished and the result URL contains actual data. Default to use client's job tags if exists.
        local_cache_path (optional): Specify the absolute download path. if the user does not provide this,
            the function will create a temporary directory.

    Returns:
        pandas DataFrame
    """
    return get_result_df(client=client, data_format=data_format, res_url=res_url, local_cache_path=local_cache_path)


def get_result_spark_df(
    spark: SparkSession,
    client: FeathrClient,
    data_format: str = None,
    res_url: str = None,
    local_cache_path: str = None,
) -> DataFrame:
    """Download the job result dataset from cloud as a Spark DataFrame.

    Args:
        spark: Spark session
        client: Feathr client
        data_format: Format to read the downloaded files. Currently support `parquet`, `delta`, `avro`, and `csv`.
            Default to use client's job tags if exists.
        res_url: Result URL to download files from. Note that this will not block the job so you need to make sure
            the job is finished and the result URL contains actual data. Default to use client's job tags if exists.
        local_cache_path (optional): Specify the absolute download path. if the user does not provide this,
            the function will create a temporary directory.

    Returns:
        Spark DataFrame
    """
    return get_result_df(
        client=client,
        data_format=data_format,
        res_url=res_url,
        local_cache_path=local_cache_path,
        spark=spark,
    )


def get_result_df(
    client: FeathrClient,
    data_format: str = None,
    res_url: str = None,
    local_cache_path: str = None,
    spark: SparkSession = None,
    format: str = None,
    is_file_path: bool = False
) -> Union[DataFrame, pd.DataFrame]:
    """Download the job result dataset from cloud as a Spark DataFrame or pandas DataFrame.

    Args:
        client: Feathr client
        data_format: Format to read the downloaded files. Currently support `parquet`, `delta`, `avro`, and `csv`.
            Default to use client's job tags if exists.
        res_url: Result URL to download files from. Note that this will not block the job so you need to make sure
            the job is finished and the result URL contains actual data. Default to use client's job tags if exists.
        local_cache_path (optional): Specify the absolute download directory. if the user does not provide this,
            the function will create a temporary directory.
        spark (optional): Spark session. If provided, the function returns spark Dataframe.
            Otherwise, it returns pd.DataFrame.
        format: An alias for `data_format` (for backward compatibility).
        is_file: If 'res_url' is a single file or a directory. Default as False

    Returns:
        Either Spark or pandas DataFrame.
    """
    if format is not None:
        data_format = format

    if data_format is None:
        # May use data format from the job tags
        if client.get_job_tags() and client.get_job_tags().get(OUTPUT_FORMAT):
            data_format = client.get_job_tags().get(OUTPUT_FORMAT)
        else:
            raise ValueError("Cannot determine the data format. Please provide the data_format argument.")

    data_format = data_format.lower()

    if is_databricks() and client.spark_runtime != "databricks":
        raise RuntimeError(f"The function is called from Databricks but the client.spark_runtime is {client.spark_runtime}.")

    # TODO Loading Synapse Delta table result into pandas has a bug: https://github.com/delta-io/delta-rs/issues/582
    if not spark and client.spark_runtime == "azure_synapse" and data_format == "delta":
        raise RuntimeError(f"Loading Delta table result from Azure Synapse into pandas DataFrame is not supported. You maybe able to use spark DataFrame to load the result instead.")

    # use a result url if it's provided by the user, otherwise use the one provided by the job
    res_url: str = res_url or client.get_job_result_uri(block=True, timeout_sec=1200)
    if res_url is None:
        raise ValueError(
            "`res_url` is None. Please make sure either you provide a res_url or make sure the job finished in FeathrClient has a valid result URI."
        )

    if client.spark_runtime == "local":
        if local_cache_path is not None:
            logger.warning(
                "In local spark mode, the result files are expected to be stored at a local storage and thus `local_cache_path` argument will be ignored."
            )
        local_cache_path = res_url

    elif client.spark_runtime == "databricks":
        if not res_url.startswith("dbfs:"):
            logger.warning(
                f"In Databricks, the result files are expected to be stored in DBFS, but the res_url {res_url} is not a dbfs path. Prefixing it with 'dbfs:/'"
            )
            res_url = f"dbfs:/{res_url.lstrip('/')}"

        if is_databricks():  # Check if the function is being called from Databricks
            if local_cache_path is not None:
                logger.warning(
                    "Result files are already in DBFS and thus `local_cache_path` will be ignored."
                )
            local_cache_path = res_url

    if local_cache_path is None:
        local_cache_path = TemporaryDirectory().name

    if local_cache_path != res_url:
        logger.info(f"{res_url} files will be downloaded into {local_cache_path}")
        client.feathr_spark_launcher.download_result(result_path=res_url, local_folder=local_cache_path, is_file_path = is_file_path)

    result_df = None
    try:
        if spark is not None:
            if data_format == "csv":
                result_df = spark.read.option("header", True).csv(local_cache_path)
            else:
                result_df = spark.read.format(data_format).load(local_cache_path)
        else:
            result_df = _load_files_to_pandas_df(
                dir_path=local_cache_path.replace("dbfs:", "/dbfs"),  # replace to python path if spark path is provided.
                data_format=data_format,
            )
    except Exception as e:
        logger.error(f"Failed to load result files from {local_cache_path} with format {data_format}.")
        raise e
    
    return result_df

def copy_cloud_dir(client: FeathrClient, source_url: str, target_url: str = None):
    source_url: str = source_url or client.get_job_result_uri(block=True, timeout_sec=1200)
    if source_url is None:
        raise RuntimeError("source_url None. Please make sure either you provide a source_url or make sure the job finished in FeathrClient has a valid result URI.")
    if target_url is None:
        raise RuntimeError("target_url None. Please make sure you provide a target_url.")

    client.feathr_spark_launcher.upload_or_get_cloud_path(source_url, target_url)
    
def cloud_dir_exists(client: FeathrClient, dir_path: str) -> bool:
    return client.feathr_spark_launcher.cloud_dir_exists(dir_path)

def _load_files_to_pandas_df(dir_path: str, data_format: str = "avro") -> pd.DataFrame:

    if data_format == "parquet":
        return pd.read_parquet(dir_path)

    elif data_format == "delta":
        from deltalake import DeltaTable
        delta = DeltaTable(dir_path)
        return delta.to_pyarrow_table().to_pandas()

    elif data_format == "avro":
        import pandavro as pdx
        if Path(dir_path).is_file():
            return pdx.read_avro(dir_path)
        else:
            try:
                return pd.concat([pdx.read_avro(f) for f in Path(dir_path).glob("*.avro")]).reset_index(drop=True)
            except ValueError:  # No object to concat when the dir is empty
                return pd.DataFrame()

    elif data_format == "csv":
        if Path(dir_path).is_file():
            return pd.read_csv(dir_path)
        else:
            try:
                return pd.concat([pd.read_csv(f) for f in Path(dir_path).glob("*.csv")]).reset_index(drop=True)
            except ValueError:  # No object to concat when the dir is empty
                return pd.DataFrame()

    else:
        raise ValueError(
            f"{data_format} is currently not supported in get_result_df. Currently only parquet, delta, avro, and csv are supported, please consider writing a customized function to read the result."
        )
            
def get_cloud_file_column_names(client: FeathrClient, path: str, format: str = "csv", is_file_path = True)->Set[str]:
    # Try to load publid cloud files without credential
    if path.startswith(("abfss:","wasbs:")):
        paths = re.split('/|@', path)
        if len(paths) < 4:
            raise RuntimeError(f"invalid cloud path: ", path)
        new_path = 'https://'+paths[3]+'/'+paths[2] + '/'
        if len(paths) > 4:
            new_path = new_path + '/'.join(paths[4:])
        if format == "csv" and is_file_path:
            try:
                df = pd.read_csv(new_path)
                return df.columns
            except:
                df = None
        # TODO: support loading other formats files
    
    try:
        df = get_result_df(client=client, data_format=format, res_url=path, is_file_path = is_file_path)
    except:
        logger.warning(f"failed to load cloud files from the path: {path} because of lack of permission or invalid path.")
        return None
    if df is None:
        logger.warning(f"failed to load cloud files from the path: {path} because of lack of permission or invalid path.")
        return None
    return df.columns



__version__ = "1.0.0"

def get_version():
    return __version__

# Decouple Feathr MAVEN Version from Feathr Python SDK Version
import os
def get_maven_artifact_fullname():
    maven_artifact_version = os.environ.get("MAVEN_ARTIFACT_VERSION", __version__)
    return f"com.linkedin.feathr:feathr_2.12:{maven_artifact_version}"


import os
from pathlib import Path

from feathr import MonitoringSettings
from feathr import MonitoringSqlSink
from test_fixture import (basic_test_setup, get_online_test_table_name)
from test_utils.constants import Constants


def test_feature_monitoring():
    monitor_sink_table = get_online_test_table_name("nycTaxiCITableMonitoring")
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"

    client = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    monitor_sink = MonitoringSqlSink(table_name=monitor_sink_table)
    settings = MonitoringSettings("monitoringSetting",
                                  sinks=[monitor_sink],
                                  feature_names=[
                                      "f_location_avg_fare", "f_location_max_fare"])
    client.monitor_features(settings)
    # just assume the job is successful without validating the actual result in Redis. Might need to consolidate
    # this part with the test_feathr_online_store test case
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)



from abc import ABC, abstractmethod


class HoconConvertible(ABC):
    """Represent classes that can convert into Feathr HOCON config.
    """
    @abstractmethod
    def to_feature_config(self) -> str:
        """Convert the feature anchor definition into internal HOCON format. (For internal use ony)"""
        pass


from collections import namedtuple
import copy
import json
import os
from os.path import basename
from pathlib import Path
import time
from typing import Dict, List, Optional, Union
from urllib.parse import urlparse
from urllib.request import urlopen

from databricks_cli.dbfs.api import DbfsApi
from databricks_cli.runs.api import RunsApi
from databricks_cli.dbfs.dbfs_path import DbfsPath
from databricks_cli.sdk.api_client import ApiClient
from loguru import logger
import requests
from requests.structures import CaseInsensitiveDict

from feathr.constants import *
from feathr.version import get_maven_artifact_fullname
from feathr.spark_provider._abc import SparkJobLauncher


class _FeathrDatabricksJobLauncher(SparkJobLauncher):
    """Class to interact with Databricks Spark cluster
    This is a light-weight databricks job runner, users should use the provided template json string to get more fine controlled environment for databricks cluster.
    For example, user can control whether to use a new cluster to run the job or not, specify the cluster ID, running frequency, node size, workder no., whether to send out failed notification email, etc.
    This runner will only fill in necessary arguments in the JSON template.

    This class will read from the provided configs string, and do the following steps.
    This default template can be overwritten by users, but users need to make sure the template is compatible with the default template. Specifically:
    1. it's a SparkJarTask (rather than other types of jobs, say NotebookTask or others). See https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--runs-submit for more details
    2. Use the Feathr Jar to run the job (hence will add an entry in `libraries` section)
    3. Will override `main_class_name` and `parameters` field in the JSON template `spark_jar_task` field
    4. will override the name of this job

    Args:
        workspace_instance_url (str): the workinstance url. Document to get workspace_instance_url: https://docs.microsoft.com/en-us/azure/databricks/workspace/workspace-details#workspace-url
        token_value (str): see here on how to get tokens: https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/authentication
        config_template (str): config template for databricks cluster. See https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--runs-submit for more details.
        databricks_work_dir (_type_, optional): databricks_work_dir must start with dbfs:/. Defaults to 'dbfs:/feathr_jobs'.
    """

    def __init__(
        self,
        workspace_instance_url: str,
        token_value: str,
        config_template: Union[str, Dict],
        databricks_work_dir: str = "dbfs:/feathr_jobs",
    ):
        # Below we will use Databricks job APIs (as well as many other APIs) to submit jobs or transfer files
        # For Job APIs, see https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs
        # for DBFS APIs, see: https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/dbfs
        self.config_template = config_template
        # remove possible trailing '/' due to wrong input format
        self.workspace_instance_url = workspace_instance_url.rstrip("/")
        self.auth_headers = CaseInsensitiveDict()

        # Authenticate the REST APIs. Documentation: https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/authentication
        self.auth_headers["Accept"] = "application/json"
        self.auth_headers["Authorization"] = f"Bearer {token_value}"
        self.databricks_work_dir = databricks_work_dir
        self.api_client = ApiClient(
            host=self.workspace_instance_url, token=token_value)

    def upload_or_get_cloud_path(self, local_path_or_cloud_src_path: str, tar_dir_path: Optional[str] = None):
        """
        Supports transferring file from an http path to cloud working storage, or upload directly from a local storage.
        or copying files from a source dbfs directory to a target dbfs directory
        """
        if local_path_or_cloud_src_path.startswith('dbfs') and tar_dir_path is not None:
            if not tar_dir_path.startswith('dbfs'):
                raise RuntimeError(
                    f"Failed to copy files from dbfs directory: {local_path_or_cloud_src_path}. {tar_dir_path} is not a valid target directory path"
                )
            if not self.cloud_dir_exists(local_path_or_cloud_src_path):
                raise RuntimeError(
                    f"Source folder:{local_path_or_cloud_src_path} doesn't exist. Please make sure it's a valid path")
            if self.cloud_dir_exists(tar_dir_path):
                logger.warning(
                    'Target cloud directory {} already exists. Please use another one.', tar_dir_path)
                return tar_dir_path
            DbfsApi(self.api_client).cp(recursive=True, overwrite=False,
                                        src=local_path_or_cloud_src_path, dst=tar_dir_path)
            logger.info('{} is copied to location: {}',
                        local_path_or_cloud_src_path, tar_dir_path)
            return tar_dir_path

        src_parse_result = urlparse(local_path_or_cloud_src_path)
        file_name = os.path.basename(local_path_or_cloud_src_path)
        # returned paths for the uploaded file. Note that we cannot use os.path.join here, since in Windows system it will yield paths like this:
        # dbfs:/feathrazure_cijob_snowflake_9_30_157692\auto_generated_derived_features.conf, where the path sep is mixed, and won't be able to be parsed by databricks.
        # so we force the path to be Linux style here.
        cloud_dest_path = self.databricks_work_dir + "/" + file_name
        if src_parse_result.scheme.startswith('http'):
            with urlopen(local_path_or_cloud_src_path) as f:
                # use REST API to avoid local temp file
                data = f.read()
                files = {"file": data}
                # for DBFS APIs, see: https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/dbfs
                r = requests.post(url=self.workspace_instance_url+'/api/2.0/dbfs/put',
                                  headers=self.auth_headers, files=files,  data={'overwrite': 'true', 'path': cloud_dest_path})
                logger.info('{} is downloaded and then uploaded to location: {}',
                            local_path_or_cloud_src_path, cloud_dest_path)
        elif src_parse_result.scheme.startswith('dbfs'):
            # passed a cloud path
            logger.info(
                'Skip uploading file {} as the file starts with dbfs:/', local_path_or_cloud_src_path)
            cloud_dest_path = local_path_or_cloud_src_path
        elif src_parse_result.scheme.startswith(('wasb', 's3', 'gs')):
            # if the path starts with a location that's not a local path
            logger.error(
                "File {} cannot be downloaded. Please upload the file to dbfs manually.", local_path_or_cloud_src_path
            )
            raise RuntimeError(
                f"File {local_path_or_cloud_src_path} cannot be downloaded. Please upload the file to dbfs manually."
            )
        else:
            # else it should be a local file path or dir
            if os.path.isdir(local_path_or_cloud_src_path):
                logger.info("Uploading folder {}",
                            local_path_or_cloud_src_path)
                dest_paths = []
                for item in Path(local_path_or_cloud_src_path).glob('**/*.conf'):
                    cloud_dest_path = self._upload_local_file_to_workspace(
                        item.resolve())
                    dest_paths.extend([cloud_dest_path])
                cloud_dest_path = ','.join(dest_paths)
            else:
                cloud_dest_path = self._upload_local_file_to_workspace(
                    local_path_or_cloud_src_path)
        return cloud_dest_path

    def _upload_local_file_to_workspace(self, local_path: str) -> str:
        """
        Supports transferring file from a local path to cloud working storage.
        """
        file_name = os.path.basename(local_path)
        # returned paths for the uploaded file. Note that we cannot use os.path.join here, since in Windows system it will yield paths like this:
        # dbfs:/feathrazure_cijob_snowflake_9_30_157692\auto_generated_derived_features.conf, where the path sep is mixed, and won't be able to be parsed by databricks.
        # so we force the path to be Linux style here.
        cloud_dest_path = self.databricks_work_dir + "/" + file_name
        # `local_path_or_http_path` will be either string or PathLib object, so normalize it to string
        local_path = str(local_path)
        try:
            DbfsApi(self.api_client).cp(recursive=True, overwrite=True,
                                        src=local_path, dst=cloud_dest_path)
        except RuntimeError as e:
            raise RuntimeError(
                f"The source path: {local_path}, or the destination path: {cloud_dest_path}, is/are not valid.") from e
        return cloud_dest_path

    def submit_feathr_job(
        self,
        job_name: str,
        main_jar_path: str,
        main_class_name: str,
        arguments: List[str],
        python_files: List[str],
        reference_files_path: List[str] = [],
        job_tags: Dict[str, str] = None,
        configuration: Dict[str, str] = {},
        properties: Dict[str, str] = {},
    ):
        """
        submit the feathr job to databricks
        Refer to the databricks doc for more details on the meaning of the parameters:
        https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--runs-submit
        Args:
            main_file_path (str): main file paths, usually your main jar file
            main_class_name (str): name of your main class
            arguments (str): all the arguments you want to pass into the spark job
            job_tags (str): tags of the job, for example you might want to put your user ID, or a tag with a certain information
            configuration (Dict[str, str]): Additional configs for the spark job
            properties (Dict[str, str]): Additional System Properties for the spark job
        """

        if properties:
            arguments.append("--system-properties=%s" % json.dumps(properties))

        if isinstance(self.config_template, str):
            # if the input is a string, load it directly
            submission_params = json.loads(self.config_template)
        else:
            # otherwise users might have missed the quotes in the config. Treat them as dict
            # Note that we need to use deep copy here, in order to make `self.config_template` immutable
            # Otherwise, since we need to change submission_params later, which will modify `self.config_template` and cause unexpected behaviors
            submission_params = copy.deepcopy(self.config_template)

        submission_params["run_name"] = job_name
        cfg = configuration.copy()
        if "existing_cluster_id" in submission_params:
            logger.info(
                "Using an existing general purpose cluster to run the feathr job...")
            if cfg:
                logger.warning(
                    "Spark execution configuration will be ignored. To use job-specific spark configs, please use a new job cluster or set the configs via Databricks UI."
                )
            if job_tags:
                logger.warning(
                    "Job tags will be ignored. To assign job tags to the cluster, please use a new job cluster."
                )
        elif "new_cluster" in submission_params:
            logger.info("Using a new job cluster to run the feathr job...")
            # if users don't specify existing_cluster_id
            # Solving this issue: Handshake fails trying to connect from Azure Databricks to Azure PostgreSQL with SSL
            # https://docs.microsoft.com/en-us/answers/questions/170730/handshake-fails-trying-to-connect-from-azure-datab.html
            cfg["spark.executor.extraJavaOptions"] = "-Djava.security.properties="
            cfg["spark.driver.extraJavaOptions"] = "-Djava.security.properties="
            submission_params["new_cluster"]["spark_conf"] = cfg

            if job_tags:
                custom_tags = submission_params["new_cluster"].get(
                    "custom_tags", {})
                for tag, value in job_tags.items():
                    custom_tags[tag] = value

                submission_params["new_cluster"]["custom_tags"] = custom_tags
        else:
            # TODO we should fail fast -- maybe check this in config verification while initializing the client.
            raise ValueError(
                "No cluster specifications are found. Either 'existing_cluster_id' or 'new_cluster' should be configured via feathr config."
            )

        # the feathr main jar file is anyway needed regardless it's pyspark or scala spark
        if not main_jar_path:
            logger.info(
                f"Main JAR file is not set, using default package '{get_maven_artifact_fullname()}' from Maven")
            submission_params['libraries'][0]['maven'] = {
                "coordinates": get_maven_artifact_fullname()}
            # Add json-schema dependency
            # TODO: find a proper way deal with unresolved dependencies
            # Since we are adding another entry to the config, make sure that the spark config passed as part of execution also contains a libraries array of atleast size 2
            # else you will get List Index Out of Bound exception
            # Example from feathr_config.yaml -
            # config_template: {"run_name":"FEATHR_FILL_IN",.....,"libraries":[{}, {}],".......}
        else:
            submission_params["libraries"][0]["jar"] = self.upload_or_get_cloud_path(
                main_jar_path)
        # see here for the submission parameter definition https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--request-structure-6
        if python_files:
            # this is a pyspark job. definition here: https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--sparkpythontask
            # the first file is the pyspark driver code. we only need the driver code to execute pyspark
            param_and_file_dict = {
                "parameters": arguments,
                "python_file": self.upload_or_get_cloud_path(python_files[0]),
            }
            # indicates this is a pyspark job
            # `setdefault` method will get the value of the "spark_python_task" item, if the "spark_python_task" item does not exist, insert "spark_python_task" with the value "param_and_file_dict":
            submission_params.setdefault(
                "spark_python_task", param_and_file_dict)
        else:
            # this is a scala spark job
            submission_params["spark_jar_task"]["parameters"] = arguments
            submission_params["spark_jar_task"]["main_class_name"] = main_class_name

        result = RunsApi(self.api_client).submit_run(submission_params)

        try:
            # see if we can parse the returned result
            self.res_job_id = result["run_id"]
        except:
            logger.error(
                "Submitting Feathr job to Databricks cluster failed. Message returned from Databricks: {}", result
            )
            exit(1)

        result = RunsApi(self.api_client).get_run(self.res_job_id)
        self.job_url = result["run_page_url"]
        logger.info(
            "Feathr job Submitted Successfully. View more details here: {}", self.job_url)

        # return ID as the submission result
        return self.res_job_id

    def wait_for_completion(self, timeout_seconds: Optional[int] = 600) -> bool:
        """Returns true if the job completed successfully"""
        start_time = time.time()
        while (timeout_seconds is None) or (time.time() - start_time < timeout_seconds):
            status = self.get_status()
            logger.debug("Current Spark job status: {}", status)
            # see all the status here:
            # https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--runlifecyclestate
            # https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--runresultstate
            if status in {"SUCCESS"}:
                return True
            elif status in {"INTERNAL_ERROR", "FAILED", "TIMEDOUT", "CANCELED"}:
                result = RunsApi(self.api_client).get_run_output(
                    self.res_job_id)
                # See here for the returned fields: https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--response-structure-8
                # print out logs and stack trace if the job has failed
                logger.error(
                    "Feathr job has failed. Please visit this page to view error message: {}", self.job_url)
                if "error" in result:
                    logger.error("Error Code: {}", result["error"])
                if "error_trace" in result:
                    logger.error("{}", result["error_trace"])
                return False
            else:
                time.sleep(30)
        else:
            raise TimeoutError("Timeout waiting for Feathr job to complete")

    def get_status(self) -> str:
        assert self.res_job_id is not None
        result = RunsApi(self.api_client).get_run(self.res_job_id)
        # first try to get result state. it might not be available, and if that's the case, try to get life_cycle_state
        # see result structure: https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--response-structure-6
        res_state = result["state"].get(
            "result_state") or result["state"]["life_cycle_state"]
        assert res_state is not None
        return res_state

    def get_job_result_uri(self) -> str:
        """Get job output uri

        Returns:
            str: `output_path` field in the job tags
        """
        custom_tags = self.get_job_tags()
        # in case users call this API even when there's no tags available
        return None if custom_tags is None else custom_tags[OUTPUT_PATH_TAG]

    def get_job_tags(self) -> Dict[str, str]:
        """Get job tags

        Returns:
            Dict[str, str]: a dict of job tags
        """
        assert self.res_job_id is not None
        # For result structure, see https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--response-structure-6
        result = RunsApi(self.api_client).get_run(self.res_job_id)

        if "new_cluster" in result["cluster_spec"]:
            custom_tags = result["cluster_spec"]["new_cluster"].get(
                "custom_tags")
            return custom_tags
        else:
            # this is not a new cluster; it's an existing cluster.
            logger.warning(
                "Job tags are not available since you are using an existing Databricks cluster. Consider using 'new_cluster' in databricks configuration."
            )
            return None

    def download_result(self, result_path: str, local_folder: str, is_file_path: bool = False):
        """
        Supports downloading files from the result folder. Only support paths starts with `dbfs:/` and only support downloading files in one folder (per Spark's design, everything will be in the result folder in a flat manner)
        """
        if not result_path.startswith("dbfs"):
            raise RuntimeError(
                'Currently only paths starting with dbfs is supported for downloading results from a databricks cluster. The path should start with "dbfs:" .'
            )

        recursive = True if not is_file_path else False
        DbfsApi(self.api_client).cp(recursive=recursive, overwrite=True, src=result_path, dst=local_folder)
        
    def cloud_dir_exists(self, dir_path: str):
        """
        Check if a directory of hdfs already exists
        """
        if not dir_path.startswith('dbfs'):
            raise RuntimeError(
                'Currently only paths starting with dbfs is supported. The paths should start with \"dbfs:\" .')

        try:
            DbfsApi(self.api_client).list_files(DbfsPath(dir_path))
            return True
        except:
            return False
